# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Iterator, Tuple


def EachAncestor(cls: type) -> Iterator[type]:
    """"""
    output = [cls]

    parents = cls.__bases__
    while parents.__len__() > 0:
        if parents.__len__() > 1:
            raise RuntimeError(
                f"{cls}: Somewhere in the hierarchy, "
                f"there are {parents.__len__()} ancestors instead of just 1; "
                f"Invalid case in present context"
            )

        parent = parents[0]
        output.append(parent)
        parents = parent.__bases__

    return reversed(output)


def AllSlotsOfClass(cls: type) -> Tuple[str, ...]:
    """"""
    output = []

    for ancestor in EachAncestor(cls):
        if hasattr(ancestor, "__slots__"):
            output.extend(ancestor.__slots__)

    return tuple(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from collections import namedtuple as namedtuple_t
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as pl_
import numpy as np_
import scipy.interpolate as in_
import skimage.measure as ms_

import skl_graph.brick.elm_id as id_
import skl_graph.brick.hierarchy as hrcy
import skl_graph.brick.topology_map as bymp
from skl_graph.brick.constants import UNTESTED_VALIDITY
from skl_graph.brick.node import branch_node_t, end_node_t, node_t
from skl_graph.brick.plot import direction_style_t, edge_styles_h, label_style_t


array_t = np_.ndarray
# ww_length=width-weighted length
# sq_lengths=squared lengths; Interest: all integers
edge_lengths_t = namedtuple_t("edge_lengths_t", "length ww_length lengths sq_lengths")


class raw_edge_t:
    #
    __slots__ = (
        "dim",
        "sites",
    )

    dim: int
    sites: Tuple[array_t, ...]

    def __init__(self):
        #
        for slot in self.__class__.__slots__:
            setattr(self, slot, None)

    @classmethod
    def NewWithSites(cls, sites: Tuple[array_t, ...]) -> raw_edge_t:
        #
        instance = cls()

        instance.dim = sites.__len__()
        instance.sites = _ReOrderedSites(sites)

        return instance

    @property
    def n_sites(self) -> int:
        #
        return self.sites[0].size

    def __str__(self) -> str:
        """"""
        return (
            f"{self.__class__.__name__}:\n"
            f"    Sites[{self.dim}-D]={self.sites[0].size}"
        )


class edge_t(raw_edge_t):
    #
    __slots__ = (
        "uid",
        "lengths",
        "widths",
        "invalidities",
        "_cache",
    )

    uid: str
    lengths: edge_lengths_t
    widths: array_t
    invalidities: List[str]  # Use brick.constants.UNTESTED_VALIDITY as initial value
    _cache: Dict[str, Any]

    def __init__(self):
        #
        super().__init__()
        self.invalidities = UNTESTED_VALIDITY
        self._cache = {}

    @classmethod
    def NewWithDetails(
        cls,
        sites: Tuple[array_t, ...],
        adjacent_node_uids: Sequence[str],
        width_map: array_t = None,
    ) -> edge_t:
        #
        raw_edge = raw_edge_t.NewWithSites(sites)
        ip_edge = _in_progress_edge_t.NewFromRaw(raw_edge)

        ip_edge.SetUID(adjacent_node_uids)
        ip_edge.AddWidths(width_map)
        ip_edge.SetLengths(ip_edge.widths)

        return ip_edge.AsEdge()

    @property
    def is_probably_valid(self) -> bool:
        """Partial validity test.

        Missing validity test: The end sites must be found in the adjacent nodes position/sites, which can only be
        tested at the graph level. See HasValidEndSites.

        Returns
        -------
        bool

        """
        output = True
        self.invalidities = []

        n_sites = self.sites[0].size
        length = self.lengths.length
        sq_lengths = self.lengths.sq_lengths
        if length < n_sites - 1:
            output = False
            self.invalidities.append(
                f"{length}: Computed length cannot be smaller "
                f"than number of sites - 1={n_sites - 1}"
            )
        if np_.any(sq_lengths == 0):
            output = False
            self.invalidities.append("Has repeated sites")
        if np_.any(sq_lengths > self.dim):
            output = False
            self.invalidities.append("Has site gaps")

        return output

    def HasValidEndSites(
        self,
        origin: Union[end_node_t, branch_node_t],
        destination: Union[end_node_t, branch_node_t],
    ) -> bool:
        """"""
        end_sites = self.end_sites
        n_found_end_sites = [0, 0]

        for node in (origin, destination):
            for e_idx in (0, 1):
                if isinstance(node, end_node_t):
                    end_site_found = np_.array_equal(end_sites[e_idx], node.position)
                else:
                    end_site = np_.reshape(end_sites[e_idx], (self.dim, 1))
                    end_site_found = any(
                        np_.all(end_site == np_.array(node.sites), axis=0)
                    )
                if end_site_found:
                    n_found_end_sites[e_idx] += 1

        if origin.uid == destination.uid:
            n_expected = 2
        else:
            n_expected = 1

        return n_found_end_sites == [n_expected, n_expected]

    @property
    def end_sites(self) -> Tuple[Tuple[Any, ...]]:
        """"""
        return tuple(
            tuple(self.sites[idx_][_s_idx] for idx_ in range(self.dim))
            for _s_idx in (0, -1)
        )

    def AsPolyline(self, max_distance: float = None) -> Tuple[float, array_t]:
        """
        Implementation of the Ramer–Douglas–Peucker algorithm
        https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm
        """
        cache_entry = self.AsPolyline.__name__

        if cache_entry not in self._cache:
            self._cache[cache_entry] = {}

        if max_distance is None:
            if self._cache[cache_entry].__len__() > 0:
                max_distance = min(self._cache[cache_entry].keys())
            else:
                max_distance = 1.0

        if max_distance not in self._cache[cache_entry]:
            output = np_.zeros(self.n_sites, dtype=np_.bool_)
            output[0] = True
            output[-1] = True

            sites_as_array = np_.array(self.sites).T
            segments = [(0, self.n_sites - 1)]
            while segments.__len__() > 0:
                first_idx, last_idx = segments.pop()
                first_point = sites_as_array[first_idx, :]
                along_line = sites_as_array[last_idx, :] - first_point
                al_norm = np_.linalg.norm(along_line)  # al=along line

                highest_distance = 0.0
                idx_of_hd = first_idx  # hd=highest distance
                for idx in range(first_idx + 1, last_idx):
                    distance = _DistanceToLine(
                        sites_as_array[idx, :], first_point, along_line, al_norm
                    )
                    if distance > highest_distance:
                        idx_of_hd = idx
                        highest_distance = distance

                if highest_distance > max_distance:
                    output[idx_of_hd] = True
                    segments.extend(((first_idx, idx_of_hd), (idx_of_hd, last_idx)))

            self._cache[cache_entry][max_distance] = (
                max_distance,
                np_.nonzero(output)[0],
            )

        return self._cache[cache_entry][max_distance]

    def AsCurve(self) -> Optional[tuple]:
        """"""
        cache_entry = self.AsCurve.__name__

        if cache_entry not in self._cache:
            if self.n_sites > 1:
                arc_lengths = np_.cumsum([0] + self.lengths.sq_lengths.tolist())
                self._cache[cache_entry] = tuple(
                    in_.PchipInterpolator(arc_lengths, self.sites[idx_])
                    for idx_ in range(self.dim)
                )
            else:
                self._cache[cache_entry] = None

        return self._cache[cache_entry]

    def OriginDirection(self) -> Optional[array_t]:
        """"""
        return self._Direction(self.OriginDirection)

    def FinalDirection(self) -> Optional[array_t]:
        """"""
        return self._Direction(self.FinalDirection)

    def _Direction(self, where: Callable) -> array_t:
        """"""
        where = where.__name__
        if where not in self._cache:
            self._SetEndPointDirections()

        return self._cache[where]

    def _SetEndPointDirections(self) -> None:
        #
        cache_entry_o = self.OriginDirection.__name__
        cache_entry_f = self.FinalDirection.__name__

        as_curve = self.AsCurve()
        if as_curve is None:
            self._cache[cache_entry_o] = None
            self._cache[cache_entry_f] = None
        else:
            max_arclength = as_curve[0].x.item(-1)
            o_dir, f_dir = [], []
            for d_idx in range(self.dim):
                directions = as_curve[d_idx]((0, max_arclength), 1)
                o_dir.append(directions[0])
                f_dir.append(directions[1])
            self._cache[cache_entry_o] = np_.array(o_dir, dtype=np_.float64) / (
                -np_.linalg.norm(o_dir)
            )
            self._cache[cache_entry_f] = np_.array(
                f_dir, dtype=np_.float64
            ) / np_.linalg.norm(f_dir)

    def __str__(self) -> str:
        """"""
        if (self._cache is None) or (self._cache.__len__() == 0):
            cached_values = "None yet"
        else:
            cached_values = ", ".join(self._cache.keys())

        return _EdgeCommonDescription(self) + f"\n    Cached values: {cached_values}"


class _in_progress_edge_t(raw_edge_t):
    #
    __slots__ = (
        "uid",
        "lengths",
        "widths",
    )

    uid: str
    lengths: edge_lengths_t
    widths: array_t

    @classmethod
    def NewFromRaw(cls, raw_edge: raw_edge_t) -> _in_progress_edge_t:
        #
        instance = cls()

        for slot in hrcy.AllSlotsOfClass(raw_edge.__class__):
            setattr(instance, slot, getattr(raw_edge, slot))

        return instance

    def AsEdge(self) -> edge_t:
        #
        output = edge_t()

        for slot in hrcy.AllSlotsOfClass(self.__class__):
            setattr(output, slot, getattr(self, slot))

        return output

    def SetUID(self, adjacent_node_uids: Sequence[str]) -> None:
        """"""
        if adjacent_node_uids.__len__() != 2:
            raise RuntimeError(
                f"{adjacent_node_uids.__len__()}: Incorrect number of adjacent node uids"
            )

        node_uid_0, node_uid_1 = adjacent_node_uids
        if node_uid_0 > node_uid_1:
            node_uid_0, node_uid_1 = node_uid_1, node_uid_0

        uid_components = [
            id_.EncodedNumber(coord)
            for coord in node_uid_0.split(id_.COORDINATE_SEPARATOR)
        ]
        uid_components.append(id_.COORDINATE_SEPARATOR)
        uid_components.extend(
            id_.EncodedNumber(coord)
            for coord in node_uid_1.split(id_.COORDINATE_SEPARATOR)
        )

        self.uid = "".join(uid_components)

    def SetLengths(self, widths: array_t) -> None:
        """
        Passing widths instead of using self.widths ensures that this method will not be called before setting
        self.widths, should it be set.
        """
        sites_as_array = np_.array(self.sites)
        segments = np_.diff(sites_as_array, axis=1)
        sq_lengths = (segments**2).sum(axis=0)
        lengths = np_.sqrt(sq_lengths)
        length = lengths.sum().item()

        if widths is None:
            ww_length = -1.0
        else:
            ww_length = (0.5 * (widths[1:] + widths[:-1]) * lengths).sum().item()

        self.lengths = edge_lengths_t(
            length=length, ww_length=ww_length, lengths=lengths, sq_lengths=sq_lengths
        )

    def AddWidths(self, width_map: array_t) -> None:
        #
        if width_map is not None:
            self.widths = width_map[self.sites]

    def AppendBranchNode(
        self,
        b_coords: array_t,
        node: node_t,
        adjacent_node_uids: List[str],
        force_after: bool = False,
    ) -> None:
        #
        adjacent_node_uids.append(node.uid)

        space_dim = self.dim
        first_site = tuple(self.sites[idx_][0] for idx_ in range(space_dim))
        sq_distance = (np_.subtract(first_site, b_coords) ** 2).sum()

        if self.n_sites > 1:
            # 0 <: so that if the edge is a self-loop ending at the same site, it does not put twice the site in a row
            if 0 < sq_distance <= space_dim:
                self.sites = tuple(
                    np_.hstack((b_coords[idx_], self.sites[idx_]))
                    for idx_ in range(space_dim)
                )
            else:
                self.sites = tuple(
                    np_.hstack((self.sites[idx_], b_coords[idx_]))
                    for idx_ in range(space_dim)
                )
        elif force_after:
            self.sites = tuple(
                np_.hstack((self.sites[idx_], b_coords[idx_]))
                for idx_ in range(space_dim)
            )
        else:
            self.sites = tuple(
                np_.hstack((b_coords[idx_], self.sites[idx_]))
                for idx_ in range(space_dim)
            )

    def __str__(self) -> str:
        """"""
        return _EdgeCommonDescription(self)


def _EdgeCommonDescription(edge: Union[edge_t, _in_progress_edge_t]) -> str:
    """"""
    origin = tuple(edge.sites[idx][0] for idx in range(edge.dim))
    if edge.lengths is None:
        raw_length = "Not computed yet"
        ww_length = raw_length
    else:
        raw_length = round(edge.lengths.length, 2)
        ww_length = round(edge.lengths.ww_length, 2)

    return (
        f"{edge.__class__.__name__}[{edge.uid}]:\n"
        f"    Sites[{edge.dim}-D]={edge.sites[0].size}\n"
        f"    Origin: {origin}\n"
        f"    Lengths: Raw={raw_length}, WW={ww_length}"
    )


def RawEdges(
    skl_map: array_t, b_node_lmap: array_t
) -> Tuple[Sequence[raw_edge_t], array_t]:
    """"""
    edge_map = skl_map.astype(np_.int8)
    edge_map[b_node_lmap > 0] = 0
    edge_lmap, n_edges = bymp.LABELING_FCT_FOR_DIM[skl_map.ndim](edge_map)

    edge_props = ms_.regionprops(edge_lmap)

    edges = n_edges * [raw_edge_t()]
    for props in edge_props:
        sites = props.image.nonzero()
        for d_idx in range(skl_map.ndim):
            sites[d_idx].__iadd__(props.bbox[d_idx])
        edges[props.label - 1] = raw_edge_t.NewWithSites(sites)

    return edges, edge_lmap


def EdgesFromRawEdges(
    raw_edges: Sequence[raw_edge_t],
    e_nodes: Sequence[end_node_t],
    b_nodes: Sequence[branch_node_t],
    edge_lmap: array_t,
    e_node_lmap: array_t,
    b_node_lmap: array_t,
    width_map: array_t = None,
) -> Tuple[Tuple[edge_t], List[List[str]]]:
    #
    edge_tmap = bymp.TopologyMapOfMap(edge_lmap > 0)
    # ip_=in progress
    ip_edges = [_in_progress_edge_t.NewFromRaw(edge) for edge in raw_edges]

    # ep=edge end point; Keep < 2 since ==0 (length-1 edges) and ==1 (other edges) are needed
    # Do not use list multiplication since the same list then used for all the elements
    node_uids_per_edge = [[] for _ in ip_edges]
    for ep_coords in zip(*(edge_tmap < 2).nonzero()):
        edge_idx = edge_lmap[ep_coords] - 1
        edge = ip_edges[edge_idx]
        e_node_label = e_node_lmap[ep_coords]

        if e_node_label > 0:
            # End node-to-X edge (i.e., edge end point is also an end node)
            node_uids_per_edge[edge_idx].append(e_nodes[e_node_label - 1].uid)
            if edge.n_sites == 1:
                # End node-to-branch node edge (and there is a unique non-zero value in b_neighborhood)
                nh_slices_starts, b_neighborhood = _LMapNeighborhood(
                    b_node_lmap, ep_coords
                )
                b_node_label = np_.amax(b_neighborhood)
                b_coords = np_.transpose((b_neighborhood == b_node_label).nonzero())[0]
                edge.AppendBranchNode(
                    np_.add(nh_slices_starts, b_coords),
                    b_nodes[b_node_label - 1],
                    node_uids_per_edge[edge_idx],
                )
        else:
            nh_slices_starts, b_neighborhood = _LMapNeighborhood(b_node_lmap, ep_coords)
            force_after = False
            # Looping only for length-1, b-to-b edges
            for b_coords in zip(*b_neighborhood.nonzero()):
                b_node_label = b_neighborhood[b_coords]
                edge.AppendBranchNode(
                    np_.add(nh_slices_starts, b_coords),
                    b_nodes[b_node_label - 1],
                    node_uids_per_edge[edge_idx],
                    force_after=force_after,
                )
                force_after = not force_after

    for edge, adjacent_node_uids in zip(ip_edges, node_uids_per_edge):
        edge.SetUID(adjacent_node_uids)
        edge.AddWidths(width_map)
        edge.SetLengths(edge.widths)

    edges = tuple(edge.AsEdge() for edge in ip_edges)

    return edges, node_uids_per_edge


def Plot(
    edges: Iterable[Tuple[str, str, edge_t]],
    transformation: Callable[[array_t], array_t],
    vector_transf: Callable[[array_t], array_t],
    axes: pl_.axes.Axes,
    edge_styles: edge_styles_h,
    direction_style: direction_style_t,
    label_style: label_style_t,
    mode: str = "site",  # "site", "polyline", "curve"
    max_distance: float = 1.0,
) -> None:
    #
    # space_dim = edges[0][2].dim  # Does not work since 'MultiEdgeDataView' object is not subscriptable
    space_dim = 2
    for _, _, edge in edges:
        space_dim = edge.dim
        break

    plot_fct = axes.plot if space_dim == 2 else axes.plot3D

    for origin, destination, edge in edges:
        if mode == "curve":
            as_curve = edge.AsCurve()
            if as_curve is None:
                sites = list(edge.sites)
            else:
                max_arc_length = as_curve[0].x.item(-1)
                step = 0.125
                arc_lengths = np_.arange(0.0, max_arc_length + 0.5 * step, step)
                sites = [as_curve[idx_](arc_lengths) for idx_ in range(space_dim)]
        elif mode == "polyline":
            polyline_idc = edge.AsPolyline(max_distance=max_distance)[1]
            sites = [edge.sites[idx_][polyline_idc] for idx_ in range(space_dim)]
        elif mode == "site":
            sites = list(edge.sites)
        else:
            raise ValueError(
                f"{mode}: Invalid plotting mode; Valid modes: 'site', 'polyline', 'curve'"
            )
        sites[0], sites[1] = sites[1], transformation(sites[0])

        if origin == destination:
            edge_style = edge_styles[1]
        else:
            edge_style = edge_styles[0]
        plot_style = f"{edge_style.color}." if mode == "site" else edge_style.color
        plot_fct(
            *sites,
            plot_style + edge_style.type,
            linewidth=edge_style.size,
            markersize=edge_style.size,
        )

        if direction_style.show:
            origin_direction = edge.OriginDirection()
            if origin_direction is not None:
                dir_sites = tuple(
                    np_.hstack((sites[idx_][0], sites[idx_][-1]))
                    for idx_ in range(space_dim)
                )
                directions = list(zip(origin_direction, edge.FinalDirection()))
                directions[0], directions[1] = (
                    directions[1],
                    vector_transf(directions[0]),
                )
                axes.quiver(
                    *dir_sites,
                    *directions,
                    color=direction_style.color,
                    linewidth=direction_style.size,
                )

        # TODO: plot edge labels


def _ReOrderedSites(sites: Tuple[array_t, ...]) -> Tuple[array_t, ...]:
    """
    If the number of sites is 1 or 2, the input argument is returned (i.e., no copy is made).

    Parameters
    ----------
    sites

    Returns
    -------

    """
    n_sites = sites[0].size
    if n_sites < 3:
        return sites

    dim = sites.__len__()

    self_loop = all(sites[idx][0] == sites[idx][-1] for idx in range(dim))
    if self_loop:
        sites = tuple(sites[idx][:-1] for idx in range(dim))
        n_sites -= 1
        self_origin = np_.fromiter(
            (sites[idx][0] for idx in range(dim)), dtype=sites[0].dtype
        )
        self_origin = np_.reshape(self_origin, (1, dim))
    else:
        self_origin = None

    sites_as_array = np_.transpose(np_.array(sites))
    reordered_coords = [np_.array([sites[idx][0] for idx in range(sites.__len__())])]
    unvisited_slc = np_.ones(n_sites, dtype=np_.bool_)
    unvisited_slc[0] = False
    unvisited_sites = None
    end_point = None
    pre_done = False
    post_done = False

    while unvisited_slc.any():
        if post_done:
            neighbor_idc = ()
        else:
            end_point = reordered_coords[-1]
            neighbor_idc, unvisited_sites = _NeighborIndices(
                dim, sites_as_array, unvisited_slc, end_point
            )

        if (neighbor_idc.__len__() == 1) or post_done:
            also_grow_first = (reordered_coords.__len__() > 1) and not pre_done
            if not post_done:
                c_idx = neighbor_idc[0]
                reordered_coords.append(unvisited_sites[c_idx, :])
                unvisited_slc[np_.where(unvisited_slc)[0][c_idx]] = False
            if also_grow_first:
                end_point = reordered_coords[0]
                neighbor_idc, unvisited_sites = _NeighborIndices(
                    dim, sites_as_array, unvisited_slc, end_point
                )
                if neighbor_idc.__len__() == 1:
                    c_idx = neighbor_idc[0]
                    reordered_coords = [unvisited_sites[c_idx, :]] + reordered_coords
                    unvisited_slc[np_.where(unvisited_slc)[0][c_idx]] = False
                elif neighbor_idc.__len__() == 0:
                    pre_done = True  # End point has been reached
                else:
                    raise RuntimeError(
                        f"{neighbor_idc.__len__()} neighbors when only 1 is expected\n"
                        f"{sites}\n{reordered_coords}\n{unvisited_slc}\n{end_point}"
                    )
        elif neighbor_idc.__len__() == 2:
            if reordered_coords.__len__() == 1:
                idx1, idx2 = neighbor_idc
                reordered_coords = [unvisited_sites[idx1, :]] + reordered_coords
                reordered_coords.append(unvisited_sites[idx2, :])
                true_map = np_.where(unvisited_slc)[0]
                unvisited_slc[true_map[idx1]] = False
                unvisited_slc[true_map[idx2]] = False
            else:
                raise RuntimeError(
                    f"2 neighbors when only 1 is expected\n"
                    f"{sites}\n{reordered_coords}\n{unvisited_slc}\n{end_point}"
                )
        elif neighbor_idc.__len__() == 0:
            post_done = True  # End point has been reached
        else:
            raise RuntimeError(
                f"{neighbor_idc.__len__()} neighbors when only 1 or 2 are expected\n"
                f"{sites}\n{reordered_coords}\n{unvisited_slc}\n{end_point}"
            )

    reordered_coords = np_.array(reordered_coords)
    if self_loop:
        self_origin_idx = np_.argwhere(
            np_.all(reordered_coords == self_origin, axis=1)
        ).item()
        if self_origin_idx > 0:
            reordered_coords = np_.roll(reordered_coords, -self_origin_idx, axis=0)
        reordered_coords = np_.vstack((reordered_coords, self_origin))
    reordered_coords = tuple(reordered_coords[:, _idx] for _idx in range(dim))

    return reordered_coords


def _NeighborIndices(
    dim: int, sites: array_t, unvisited_slc: array_t, end_point: array_t
) -> Tuple[array_t, array_t]:
    """"""
    unvisited_sites = sites[unvisited_slc, :]

    distances = np_.fabs(unvisited_sites - np_.reshape(end_point, (1, dim)))
    neighbor_idc = np_.nonzero(np_.all(distances <= 1, axis=1))[0]

    return neighbor_idc, unvisited_sites


def _DistanceToLine(
    point: array_t, on_line: array_t, along_line: array_t, al_norm: float
) -> float:
    #
    if al_norm == 0.0:
        return np_.linalg.norm(point - on_line)

    return np_.linalg.norm(np_.cross(point - on_line, along_line)) / al_norm


def _LMapNeighborhood(lmap: array_t, site: Tuple[int, ...]) -> Tuple[array_t, array_t]:
    #
    slices_starts = tuple(max(site[idx_] - 1, 0) for idx_ in range(site.__len__()))
    slices = tuple(
        slice(slices_starts[idx_], min(site[idx_] + 2, lmap.shape[idx_]))
        for idx_ in range(site.__len__())
    )
    neighborhood = lmap[slices]

    return np_.array(slices_starts, dtype=np_.int64), neighborhood



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
A map is a Numpy ndarray representing one or several object on a background. It has several variants. The ones concerned
here are:
- Binary   map (bmp): ndarray of type int8  where the object(s) are labeled 1 and the background is 0.
- Topology map (tmp): ndarray of type int8  where each object site (pixel, voxel...) has a value between 0
and 3**D - 1, where D is the number of dimensions of the array, and the background is 3**D. The values correspond to
the number of neighboring sites in the object with the weakest connectivity (8 in 2-D, 26 in 3-D...).

It is implicit that all the functions below require the map(s) to be valid (i.e., follow(s) the definitions above).
"""

from typing import Callable, Tuple, Union

import numpy as np_
import scipy.ndimage as im_


array_t = np_.array
labeling_fct_t = Callable[[array_t], Tuple[array_t, int]]


_SQUARE_3x3 = np_.ones((3, 3), dtype=np_.uint8)
_SQUARE_3x3x3 = np_.ones((3, 3, 3), dtype=np_.uint8)
_LABELING_FCT_2D: labeling_fct_t = lambda a_map: im_.label(
    a_map, structure=_SQUARE_3x3, output=np_.int64
)
_LABELING_FCT_3D: labeling_fct_t = lambda a_map: im_.label(
    a_map, structure=_SQUARE_3x3x3, output=np_.int64
)
LABELING_FCT_FOR_DIM = (None, None, _LABELING_FCT_2D, _LABELING_FCT_3D)


_FULL_SHIFTS_FOR_2D_NEIGHBORS = tuple(
    (i, j) for i in (-1, 0, 1) for j in (-1, 0, 1) if i != 0 or j != 0
)
_FULL_SHIFTS_FOR_3D_NEIGHBORS = tuple(
    (i, j, k)
    for i in (-1, 0, 1)
    for j in (-1, 0, 1)
    for k in (-1, 0, 1)
    if i != 0 or j != 0 or k != 0
)
_FULL_SHIFTS_FOR_NEIGHBORS_FOR_DIM = (
    None,
    None,
    _FULL_SHIFTS_FOR_2D_NEIGHBORS,
    _FULL_SHIFTS_FOR_3D_NEIGHBORS,
)

_MIN_SHIFTS_FOR_2D_NEIGHBORS = tuple(
    elm for elm in _FULL_SHIFTS_FOR_2D_NEIGHBORS if np_.abs(elm).sum() == 1
)
_MIN_SHIFTS_FOR_3D_NEIGHBORS = tuple(
    elm for elm in _FULL_SHIFTS_FOR_3D_NEIGHBORS if np_.abs(elm).sum() == 1
)
_MIN_SHIFTS_FOR_NEIGHBORS_FOR_DIM = (
    None,
    None,
    _MIN_SHIFTS_FOR_2D_NEIGHBORS,
    _MIN_SHIFTS_FOR_3D_NEIGHBORS,
)


def TopologyMapOfMap(
    a_map: array_t, full_connectivity: bool = True, return_bg_label: bool = False
) -> Union[array_t, Tuple[array_t, int]]:
    #
    # The topology map is labeled as follows: background=TMapBackgroundLabel(a_map); Pixels of the objects=number of
    # neighboring pixels that belong to the given object (as expected, isolated pixels receive 0).
    #
    # Output dtype is int instead of uint to allow for subtraction
    #
    # Works for multi-object maps.
    #
    # Note: using a_map avoids shadowing Python's map.
    #
    output = np_.array(a_map, dtype=np_.int8)

    if full_connectivity:
        shifts_for_dim = _FULL_SHIFTS_FOR_NEIGHBORS_FOR_DIM
    else:
        shifts_for_dim = _MIN_SHIFTS_FOR_NEIGHBORS_FOR_DIM
    padded_sm = np_.pad(a_map, 1, "constant")
    unpadding_domain = a_map.ndim * (slice(1, -1),)
    rolling_axes = tuple(range(a_map.ndim))
    for shifts in shifts_for_dim[a_map.ndim]:
        output += np_.roll(padded_sm, shifts, axis=rolling_axes)[unpadding_domain]

    background_label = TMapBackgroundLabel(a_map)
    output[a_map == 0] = background_label + 1
    output -= 1

    if return_bg_label:
        return output, background_label
    return output


def TMapBackgroundLabel(a_map: array_t) -> int:
    """
    Must be equal to the max number of neighbors in a map, + 1.
    Note: using a_map avoids shadowing Python's map.
    """
    return 3 ** a_map.ndim



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from typing import Callable, Dict, Iterable, List, Tuple

import matplotlib.pyplot as pl_
import numpy as np_
import scipy.spatial as sp_
import skimage.measure as ms_
from mpl_toolkits.mplot3d.art3d import Poly3DCollection as polygons_3d_t

import skl_graph.brick.elm_id as id_
import skl_graph.brick.topology_map as bymp
from skl_graph.brick.constants import UNTESTED_VALIDITY
from skl_graph.brick.plot import label_style_t, node_style_h, node_styles_h


array_t = np_.ndarray


class node_t:
    #
    __slots__ = ("uid", "position", "invalidities")

    uid: str
    position: array_t
    invalidities: List[str]  # Use brick.constants.UNTESTED_VALIDITY as initial value

    def __init__(self):
        #
        for slot in self.__class__.__slots__:
            setattr(self, slot, None)
        self.invalidities = UNTESTED_VALIDITY

    def _SetUID(self) -> None:
        #
        self.uid = id_.COORDINATE_SEPARATOR.join(
            coord.__str__() for coord in self.position
        )

    @property
    def is_valid(self) -> bool:
        """"""
        output = True
        self.invalidities = []

        if self.uid is None:
            output = False
            self.invalidities.append("Has no UID")

        return output

    def __str__(self) -> str:
        """"""
        return f"{self.__class__.__name__}[{self.uid}]: Pos={self.position}"


class end_node_t(node_t):
    #
    __slots__ = ("diameter",)

    diameter: float

    def __init__(self):
        #
        super().__init__()
        for slot in self.__class__.__slots__:
            setattr(self, slot, None)

    @classmethod
    def WithPosition(cls, position: array_t, width_map: array_t = None) -> end_node_t:
        #
        instance = cls()

        instance.position = position
        if width_map is not None:
            instance.diameter = width_map.item(tuple(position))
        instance._SetUID()

        return instance


class branch_node_t(node_t):
    #
    __slots__ = ("sites", "diameters")

    sites: Tuple[array_t, ...]
    diameters: array_t

    def __init__(self):
        #
        super().__init__()
        for slot in self.__class__.__slots__:
            setattr(self, slot, None)

    @classmethod
    def WithCentroidAndSites(
        cls, centroid: array_t, sites: Tuple[array_t, ...], width_map: array_t = None
    ) -> branch_node_t:
        #
        # TODO: why passing the centroid instead of computing it here from sites?
        # TODO: sites is not optional but seems to be considered as such from at least a calling function (do a search to find out which)
        instance = cls()

        sites_as_array = np_.array(sites)
        centroid = np_.around(centroid).reshape((-1, 1))
        segments = sites_as_array - centroid
        medoid_idx = (segments**2).sum(axis=0).argmin()
        # np_.array(): fresh array instead of a view of sites_as_array
        position = np_.array(sites_as_array[:, medoid_idx].squeeze())

        instance.position = position
        instance.sites = sites
        if width_map is not None:
            instance.diameters = width_map[sites]
        instance._SetUID()

        return instance


def EndNodes(
    part_map: array_t, width_map: array_t = None
) -> Tuple[List[end_node_t], array_t]:
    #
    # Note: End nodes are necessarily single-pixel nodes. Hence, they have no coordinate list.
    #
    # Not uint to allow for subtraction
    e_node_lmap = (part_map == 1).astype(np_.int64)  # Not really an lmsk here
    e_node_coords = np_.where(e_node_lmap)

    e_nodes = e_node_coords[0].__len__() * [end_node_t()]
    for n_idx, position in enumerate(zip(*e_node_coords)):
        end_node = end_node_t.WithPosition(
            np_.array(position, dtype=np_.int64), width_map=width_map
        )
        e_nodes[n_idx] = end_node
        e_node_lmap[position] = n_idx + 1  # Now that's becoming an lmsk

    return e_nodes, e_node_lmap


def BranchNodes(
    part_map: array_t, width_map: array_t = None
) -> Tuple[List[branch_node_t], array_t]:
    #
    # Note: Branch nodes always have a coordinate list (i.e., even if they are single-pixeled)
    #
    b_node_map = np_.logical_and(
        part_map > 2, part_map != bymp.TMapBackgroundLabel(part_map)
    )
    b_node_lmap, n_b_nodes = bymp.LABELING_FCT_FOR_DIM[part_map.ndim](b_node_map)

    b_node_props = ms_.regionprops(b_node_lmap)

    b_nodes = n_b_nodes * [branch_node_t()]
    for n_idx, props in enumerate(b_node_props):
        sites = props.image.nonzero()
        for d_idx in range(part_map.ndim):
            sites[d_idx].__iadd__(props.bbox[d_idx])
        branch_node = branch_node_t.WithCentroidAndSites(
            props.centroid, sites=sites, width_map=width_map
        )
        b_nodes[n_idx] = branch_node

    return b_nodes, b_node_lmap


def PlotEndNodes(
    nodes: Iterable[Tuple[str, node_t]],
    transformation: Callable[[array_t], array_t],
    axes: pl_.axes.Axes,
    node_style: node_style_h,
) -> None:
    """
    nodes: all the nodes of the graph, so that filtering is needed here
    """
    positions = np_.array(
        tuple(node.position for _, node in nodes if isinstance(node, end_node_t))
    )
    if positions.size == 0:
        return

    plot_style = node_style.color + node_style.type

    if positions.shape[1] == 2:
        axes.plot(
            positions[:, 1],
            transformation(positions[:, 0]),
            plot_style,
            markersize=node_style.size,
        )
    else:
        axes.plot3D(
            positions[:, 1],
            transformation(positions[:, 0]),
            positions[:, 2],
            plot_style,
            markersize=node_style.size,
        )


def Plot2DBranchNodes(
    nodes: Iterable[Tuple[str, node_t]],
    degrees: Iterable[int],
    transformation: Callable[[array_t], array_t],
    axes: pl_.axes.Axes,
    node_styles: node_styles_h,
) -> None:
    """
    nodes: all the nodes of the graph, so that filtering is needed here
    """
    positions_0, positions_1, styles, sizes = [], [], [], []
    for (_, node), degree in zip(nodes, degrees):
        degree: Tuple[str, int]
        if isinstance(node, branch_node_t):
            node_style = node_styles.get(degree[1], node_styles[None])

            coords_0 = node.sites[1]
            coords_1 = transformation(node.sites[0])
            if coords_0.size > 2:
                try:
                    hull = sp_.ConvexHull(np_.transpose((coords_0, coords_1)))
                except sp_.QhullError:
                    # TODO: check when this happens, in particular flat convex hull
                    axes.plot(
                        coords_0,
                        coords_1,
                        node_style.color + "-",
                        linewidth=0.3 * node_style.size,
                    )
                else:
                    vertices = hull.vertices
                    axes.fill(coords_0[vertices], coords_1[vertices], node_style.color)
            elif coords_0.size > 1:
                axes.plot(
                    coords_0,
                    coords_1,
                    node_style.color + "-",
                    linewidth=0.3 * node_style.size,
                )

            # Grouping for "better performances"
            positions_0.extend(coords_0)
            positions_1.extend(coords_1)
            styles.extend(coords_0.size * (node_style.color + node_style.type,))
            sizes.extend(coords_0.size * (node_style.size,))

    if positions_0.__len__() > 0:
        # NOTE: axes.plot does not accept one style and one size per node. Hence the "styles[0]" and "sizes[0]".
        axes.plot(positions_0, positions_1, styles[0], markersize=sizes[0])


def Plot3DBranchNodes(
    nodes: Iterable[Tuple[str, node_t]],
    degrees: Iterable[int],
    transformation: Callable[[array_t], array_t],
    axes: pl_.axes.Axes,
    node_styles: node_styles_h,
) -> None:
    """
    nodes: all the nodes of the graph, so that filtering is needed here
    """
    positions_0, positions_1, positions_2, styles, sizes = [], [], [], [], []
    for (_, node), degree in zip(nodes, degrees):
        degree: Tuple[str, int]
        if isinstance(node, branch_node_t):
            node_style = node_styles.get(degree[1], node_styles[None])

            coords_0 = node.sites[1]
            coords_1 = transformation(node.sites[0])
            coords_2 = node.sites[2]
            if coords_0.size > 3:
                try:
                    coords = np_.transpose((coords_0, coords_1, coords_2))
                    hull = sp_.ConvexHull(coords)
                    triangle_lst = []
                    for face in hull.simplices:
                        triangle_lst.append(
                            [coords[v_idx, :].tolist() for v_idx in face]
                        )
                    triangle_lst = polygons_3d_t(
                        triangle_lst,
                        facecolors=node_style.color,
                        edgecolors="b",
                        linewidth=0.6 * node_style.size,
                    )
                    axes.add_collection3d(triangle_lst)
                except:
                    # TODO: better space-filling drawing: to be done
                    axes.plot3D(
                        coords_0,
                        coords_1,
                        coords_2,
                        node_style.color + ".",
                        markersize=node_style.size,
                    )
            elif coords_0.size > 2:
                triangle = list(zip(coords_0, coords_1, coords_2))
                triangle_lst = polygons_3d_t([triangle], facecolors=node_style.color)
                axes.add_collection3d(triangle_lst)
            elif coords_0.size > 1:
                axes.plot3D(
                    coords_0,
                    coords_1,
                    coords_2,
                    node_style.color + "-",
                    linewidth=0.3 * node_style.size,
                )

            # Grouping for "better performances"
            positions_0.extend(coords_0)
            positions_1.extend(coords_1)
            positions_2.extend(coords_2)
            styles.extend(coords_0.size * (node_style.color + node_style.type,))
            sizes.extend(coords_0.size * (node_style.size,))

    if positions_0.__len__() > 0:
        # NOTE: axes.plot3D does not accept one style and one size per node. Hence the "styles[0]" and "sizes[0]".
        axes.plot3D(
            positions_0, positions_1, positions_2, styles[0], markersize=sizes[0]
        )


def Plot3DNodeLabels(
    nodes: Iterable[str],
    positions_as_dict: Dict[str, Tuple[int, ...]],
    axes: pl_.axes.Axes,
    style: label_style_t,
) -> None:
    #
    for node in nodes:
        axes.text(
            *positions_as_dict[node], node, fontsize=style.size, color=style.color
        )



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# skl_fgraph=Skeleton graph with computable features; Derived from base skeleton graph.

from typing import Callable, Iterable, List, Optional, Tuple  # , SupportsFloat

import numpy as np_

from skl_graph.skl_graph import skl_graph_t as skl_nfgraph_t  # nf=no feature


class skl_graph_t(skl_nfgraph_t):
    #
    @property
    def highest_degree(self) -> int:
        return max(degree for ___, degree in self.degree)

    @property
    def highest_degree_w_nodes(self) -> Tuple[int, List[str]]:
        #
        max_degree = -1
        at_nodes = None
        for node, degree in self.degree:
            if degree > max_degree:
                max_degree = degree
                at_nodes = [node]
            elif degree == max_degree:
                at_nodes.append(node)

        return max_degree, at_nodes

    @property
    def edge_lengths(self) -> Tuple[float, ...]:
        #
        return tuple(
            edge.lengths.length for ___, ___, edge in self.edges.data("as_edge_t")
        )

    @property
    def length(self) -> float:
        #
        return sum(self.edge_lengths)

    @property
    def area_as_rw_x_l(self) -> Optional[float]:
        """Area as reduced width times length.

        Returns
        -------

        """
        if self.has_widths:
            return self.ReducedWidth() * self.length
        else:
            return None

    @property
    def edge_ww_lengths(self) -> Tuple[float, ...]:
        #
        return tuple(
            edge.lengths.ww_length for ___, ___, edge in self.edges.data("as_edge_t")
        )

    @property
    def ww_length(self) -> float:
        #
        return sum(self.edge_ww_lengths)

    def EdgeReducedWidths(
        self, reduce_fct: Callable[[Iterable[float]], float] = np_.mean
    ) -> Tuple[float, ...]:
        #
        return tuple(
            reduce_fct(edge.widths) for ___, ___, edge in self.edges.data("as_edge_t")
        )

    def ReducedWidth(
        self, reduce_fct: Callable[[Iterable[float]], float] = np_.mean
    ) -> float:
        #
        all_widths = []
        for ___, ___, edge in self.edges.data("as_edge_t"):
            all_widths.extend(edge.widths)

        return reduce_fct(all_widths)

    def HeterogeneousReducedWidth(
        self,
        edge_reduce_fct: Callable[[Iterable[float]], float] = np_.mean,
        final_reduce_fct: Callable[[Iterable[float]], float] = np_.mean,
    ) -> float:
        #
        return final_reduce_fct(self.EdgeReducedWidths(edge_reduce_fct))

    def __str__(self) -> str:
        """"""
        output = super().__str__()
        output += (
            f"\n"
            f"    Highest degree={self.highest_degree}\n\n"
            f"    Length={round(self.length, 2)}"
        )
        if self.has_widths:
            output += (
                f"\n"
                f"    Width: Hom={round(self.ReducedWidth(), 2)}, "
                f"Het={round(self.HeterogeneousReducedWidth(), 2)}\n"
                f"    Area: RWxL={round(self.area_as_rw_x_l, 2)}, "
                f"WWL={round(self.ww_length, 2)}"
            )

        return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from enum import Enum as enum_t
from typing import Dict
from typing import NamedTuple as named_tuple_t
from typing import Tuple, Union

import matplotlib.pyplot as pypl


axes_t = pypl.Axes
figure_t = pypl.Figure


class plot_mode_e(enum_t):
    Networkx = 0
    SKL = 1
    SKL_Polyline = 2
    SKL_Curve = 3
    Graphviz = 4


class plot_style_t(named_tuple_t):
    type: str  # E.g., "." for nodes and "-" for edges
    size: int  # "markersize" for nodes, "linewidth" for edges
    color: str  # E.g., "b"


class direction_style_t(named_tuple_t):
    show: bool
    size: int  # "linewidth"
    color: str  # "k"


class label_style_t(named_tuple_t):
    show: bool
    size: float  # "fontsize"
    color: str  # "k"


node_style_h = plot_style_t
node_styles_h = Dict[Union[int, None], node_style_h]
node_style_unstructured_h = Tuple[str, int, str]  # type, size, color
node_styles_unstructured_h = Dict[Union[int, None], node_style_unstructured_h]

edge_style_h = plot_style_t
edge_styles_h = Tuple[edge_style_h, edge_style_h]  # regular edges, self-loops
edge_style_unstructured_h = node_style_unstructured_h
edge_styles_unstructured_h = Tuple[edge_style_unstructured_h, edge_style_unstructured_h]

direction_style_unstructured_h = Tuple[bool, int, str]  # show, size, color

label_styles_h = Tuple[label_style_t, label_style_t]  # nodes, edges
label_style_unstructured_h = Tuple[bool, float, str]  # show, size, color
label_styles_unstructured_h = Tuple[label_style_unstructured_h, label_style_unstructured_h]



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Base Skeleton Graph.

Simple example usage:
>>> import matplotlib.pyplot as pyplot
>>> import numpy
>>> import skimage.data as data
>>> import skimage.util as util
>>> object_map = util.invert(data.horse())
>>> # --- SKL Map
>>> from skl_graph.skl_map import SKLMapFromObjectMap
>>> skl_map = SKLMapFromObjectMap(object_map)
>>> # --- SKL Graph
>>> from skl_graph.skl_graph import skl_graph_t
>>> skl_graph = skl_graph_t.FromSKLMap(skl_map)
>>> # --- End of SKLGraph-based sections
>>> _, axes = pyplot.subplots(ncols=3)
>>> axes[0].matshow(object_map, cmap="gray")
>>> axes[1].matshow(skl_map, cmap="gray")
>>> skl_graph.Plot(axes=axes[2], should_block=False)
>>> for ax, title in zip(axes, ("Object", "Skeleton", "Graph")):
>>>     ax.set_title(title)
>>>     ax.set_axis_off()
>>> pyplot.tight_layout()
>>> pyplot.show()
"""

from __future__ import annotations

from typing import Callable, ClassVar, Dict, List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as pl_
import networkx as nx_
import numpy as np_
import scipy.ndimage as im_
import skimage.draw as dw_
from mpl_toolkits import mplot3d as m3_

import skl_graph.brick.edge as dg_
import skl_graph.brick.node as nd_
import skl_graph.brick.topology_map as bymp
from skl_graph.brick.constants import UNTESTED_VALIDITY
from skl_graph.brick.plot import (
    axes_t,
    direction_style_t,
    direction_style_unstructured_h,
    edge_style_unstructured_h,
    edge_styles_h,
    edge_styles_unstructured_h,
    figure_t,
    label_style_t,
    label_style_unstructured_h,
    label_styles_h,
    label_styles_unstructured_h,
    node_style_unstructured_h,
    node_styles_h,
    node_styles_unstructured_h,
    plot_mode_e,
    plot_style_t,
)


array_t = np_.ndarray


class skl_graph_t(nx_.MultiGraph):
    """
    s_node: Singleton node
    e_node: End node
    b_node: Branch node
    """

    __slots__ = (
        "dim",
        "domain_lengths",
        "n_components",
        "n_s_nodes",
        "n_e_nodes",
        "n_b_nodes",
        "has_widths",
        "invalidities",
        "node_styles",
        "edge_styles",
        "direction_style",
        "label_styles",
    )

    # TODO: use styles instead
    # TODO: use styles in all functions, not just plotexplicity (partially done for networkx)
    colormap: ClassVar[Dict[int, str]] = {0: "b", 1: "r", 2: "g"}

    dim: int
    domain_lengths: Tuple[int, ...]
    n_components: int
    n_s_nodes: int
    n_e_nodes: int
    n_b_nodes: int
    has_widths: bool
    invalidities: List[str]  # Use brick.constants.UNTESTED_VALIDITY as initial value
    node_styles: node_styles_h  # One style per degree + one default style
    edge_styles: edge_styles_h  # One style for "regular" edges, one for self-loops
    direction_style: direction_style_t
    label_styles: label_styles_h  # For nodes and edges, in that order

    def __init__(self):
        #
        super().__init__()
        for slot in self.__class__.__slots__:
            setattr(self, slot, None)
        self.invalidities = UNTESTED_VALIDITY

    @classmethod
    def FromSKLMap(cls, skl_map: array_t, width_map: array_t = None) -> skl_graph_t:
        """

        Parameters
        ----------
        skl_map : numpy.ndarray
        width_map : numpy.ndarray, optional

        Returns
        -------
        skl_graph_t
            Graph of the skeleton as an extended networkx.MultiGraph instance
        """
        instance = cls()

        instance.domain_lengths = skl_map.shape
        instance.dim = instance.domain_lengths.__len__()
        instance.has_widths = width_map is not None
        instance.n_s_nodes = 0
        instance.n_e_nodes = 0
        instance.n_b_nodes = 0

        tmap, background_label = bymp.TopologyMapOfMap(skl_map, return_bg_label=True)
        cc_map, n_components = bymp.LABELING_FCT_FOR_DIM[skl_map.ndim](skl_map)

        instance.n_components = n_components

        # Process skl_map/tmap per connected component (*)
        for cmp_label in range(1, n_components + 1):
            if n_components > 1:
                cmp_map = cc_map == cmp_label
                single_tmap = np_.full_like(tmap, background_label)
                single_tmap[cmp_map] = tmap[cmp_map]
            else:
                cmp_map = skl_map
                single_tmap = tmap

            if instance._DealsWithSpecialCases(
                single_tmap, background_label, width_map=width_map
            ):
                pass
            else:
                e_nodes, e_node_lmap = nd_.EndNodes(single_tmap, width_map=width_map)
                b_nodes, b_node_lmap = nd_.BranchNodes(single_tmap, width_map=width_map)
                raw_edges, edge_lmap = dg_.RawEdges(cmp_map, b_node_lmap)
                edges, node_uids_per_edge = dg_.EdgesFromRawEdges(
                    raw_edges,
                    e_nodes,
                    b_nodes,
                    edge_lmap,
                    e_node_lmap,
                    b_node_lmap,
                    width_map=width_map,
                )

                instance.add_nodes_from(
                    (node.uid, dict(as_node_t=node)) for node in e_nodes
                )
                instance.add_nodes_from(
                    (node.uid, dict(as_node_t=node)) for node in b_nodes
                )
                for edge, adjacent_node_uids in zip(edges, node_uids_per_edge):
                    instance._AddEdge(edge, adjacent_node_uids)

                instance.n_e_nodes += e_nodes.__len__()
                instance.n_b_nodes += b_nodes.__len__()

        instance.node_styles = {
            None: plot_style_t(type=".", size=6, color="g"),
            1: plot_style_t(type=".", size=6, color="r"),
        }
        instance.edge_styles = (
            plot_style_t(type="-", size=2, color="k"),
            plot_style_t(type=":", size=2, color="k"),
        )
        instance.direction_style = direction_style_t(show=False, size=0, color="k")
        instance.label_styles = (
            label_style_t(show=True, size=6.0, color="k"),
            label_style_t(show=False, size=6.0, color="k"),
        )

        return instance

    def _DealsWithSpecialCases(
        self, tmap: array_t, background_label: int, width_map: array_t = None
    ) -> bool:
        """Creates and adds nodes and edges of cases such as a singleton node, self loops...

        Parameters
        ----------
        tmap : numpy.ndarray
            Topological map of the skeleton; Must contain a unique connected component.
        background_label
        width_map

        Returns
        -------

        """
        singleton = np_.where(tmap == 0)
        if singleton[0].size > 0:
            # Can only be 1 since tmap is processed per connected components (*)
            singleton = np_.array(singleton, dtype=np_.int64).squeeze()
            end_node = nd_.end_node_t.WithPosition(singleton, width_map=width_map)
            self.add_node(end_node.uid, as_node_t=end_node)

            self.n_s_nodes += 1

            return True
        #
        elif (tmap[tmap != background_label] == 2).all():
            # The tmap represents a self loop
            loop_slc = np_.nonzero(tmap == 2)

            # Takes the first pixel to serve as a node, and the rest for the self-loop edge
            #
            # 0:1 makes sites elements array_t's (instead of numpy numbers), which is necessary for
            # nd_.branch_node_t.WithCentroidAndSites, but requires to squeeze the centroid.
            sites = tuple(per_dim[0:1] for per_dim in loop_slc)
            centroid = np_.array(sites, dtype=np_.float64).squeeze()
            node = nd_.branch_node_t.WithCentroidAndSites(
                centroid, sites=sites, width_map=width_map
            )
            self.add_node(node.uid, as_node_t=node)

            n_unique_sites = loop_slc[0].__len__()
            sites = tuple(
                per_dim[list(range(n_unique_sites)) + [0]] for per_dim in loop_slc
            )
            adjacent_node_uids = (node.uid, node.uid)
            edge = dg_.edge_t.NewWithDetails(
                # sites, adjacent_node_uids, node.uid, width_map=width_map
                sites,
                adjacent_node_uids,
                width_map=width_map,
            )
            self._AddEdge(edge, adjacent_node_uids)

            self.n_b_nodes += 1

            return True

        return False

    def _AddEdge(self, edge: dg_.edge_t, adjacent_node_uids: Sequence[str]) -> None:
        #
        edge_uid = edge.uid
        version_number = 1
        uid_w_vn = edge_uid
        while self.has_edge(*adjacent_node_uids, key=uid_w_vn):
            version_number += 1
            uid_w_vn = edge_uid + "+" + version_number.__str__()

        self.add_edge(*adjacent_node_uids, key=uid_w_vn, as_edge_t=edge)

    @property
    def n_nodes(self) -> int:
        return self.number_of_nodes()

    @property
    def n_edges(self) -> int:
        return self.number_of_edges()

    @property
    def is_valid(self) -> bool:
        """"""
        output = True
        self.invalidities = []

        n_components = nx_.number_connected_components(self)
        if n_components != self.n_components:
            output = False
            self.invalidities.append(
                f"Actual and stored number of connected components differ: "
                f"{n_components}!={self.n_components}"
            )
        n_s_nodes = sum(degree == 0 for _, degree in self.degree)
        if n_s_nodes != self.n_s_nodes:
            output = False
            self.invalidities.append(
                f"Actual and stored number of singleton nodes differ: "
                f"{n_s_nodes}!={self.n_s_nodes}"
            )
        n_e_nodes = sum(degree == 1 for _, degree in self.degree)
        if n_e_nodes != self.n_e_nodes:
            output = False
            self.invalidities.append(
                f"Actual and stored number of end nodes differ: "
                f"{n_e_nodes}!={self.n_e_nodes}"
            )
        n_b_nodes = sum(degree > 1 for _, degree in self.degree)
        if n_b_nodes != self.n_b_nodes:
            output = False
            self.invalidities.append(
                f"Actual and stored number of end nodes differ: "
                f"{n_b_nodes}!={self.n_b_nodes}"
            )

        nodes_are_valid = all(
            _node.is_valid for _, _node in self.nodes.data("as_node_t")
        )
        if not nodes_are_valid:
            output = False
            self.invalidities.append("Some nodes are invalid")

        some_edges_are_invalid = False
        for origin, destination, edge in self.edges.data("as_edge_t"):
            if not edge.is_probably_valid:
                output = False
                some_edges_are_invalid = True

            nodes = (self.nodes[_uid]["as_node_t"] for _uid in (origin, destination))
            if not edge.HasValidEndSites(*nodes):
                output = False
                some_edges_are_invalid = True
                edge.invalidities.append(f"Invalid sites end points")
        if some_edges_are_invalid:
            self.invalidities.append("Some edges are invalid")

        return output

    def ShowInvalidities(self) -> None:
        """"""
        if self.invalidities == UNTESTED_VALIDITY:
            _ = self.is_valid

        if self.invalidities.__len__() == 0:
            print(f"{self}\n---> Valid")
            return

        invalidities = "\n    ".join(self.invalidities)
        print(f"{self}\n---> Invalidities:\n    {invalidities}")

        for name, elements in zip(("node", "edge"), (self.nodes, self.edges)):
            print(f"    --- Invalid {name}s")
            for *_, element in elements.data(f"as_{name}_t"):
                # Since the graph has been checked, then element.invalidities != UNTESTED_VALIDITY
                if element.invalidities.__len__() > 0:
                    invalidities = "\n    ".join(element.invalidities)
                    print(f"    {element}\n    /!\\\n    {invalidities}")

    def Correctness(self, skl_map: array_t) -> Tuple[bool, bool]:
        """"""
        rebuilt_skl_map = self.RebuiltSkeletonMap()

        topological_map = bymp.TopologyMapOfMap(skl_map)
        # Keep next line before its next one
        topological_map[topological_map == bymp.TMapBackgroundLabel(skl_map)] = 0
        topological_map[topological_map > 3] = 3

        binary_correctness = np_.array_equal(rebuilt_skl_map > 0, skl_map)
        topology_correctness = np_.array_equal(rebuilt_skl_map, topological_map)

        return binary_correctness, topology_correctness

    def RebuiltSkeletonMap(self, with_width: bool = False) -> array_t:
        #
        if (not self.has_widths) and with_width:
            with_width = False
        if with_width:
            dtype = np_.float64
        else:
            # Not uint to allow for subtraction
            dtype = np_.int8

        output = np_.zeros(self.domain_lengths, dtype=dtype)

        for ___, ___, edge in self.edges.data("as_edge_t"):
            if with_width:
                output[edge.sites] = edge.widths
            else:
                output[edge.sites] = 2

        for ___, node in self.nodes.data("as_node_t"):
            if isinstance(node, nd_.branch_node_t):
                if with_width:
                    output[node.sites] = node.diameters
                else:
                    output[node.sites] = 3
            else:
                if with_width:
                    output[tuple(node.position)] = node.diameter
                else:
                    output[tuple(node.position)] = 1

        return output

    def RebuiltObjectMap(self) -> array_t:
        #
        if not self.has_widths:
            raise ValueError("Requires an SKL graph with widths")

        # Not uint to allow for subtraction
        output = np_.zeros(self.domain_lengths, dtype=np_.int8)

        if self.dim == 2:
            NewBall = dw_.disk
        else:
            NewBall = _Ball3D

        for ___, node in self.nodes.data("as_node_t"):
            if isinstance(node, nd_.branch_node_t):
                for *sites, radius in zip(
                    *node.sites,
                    np_.around(0.5 * (node.diameters - 1.0)).astype(np_.int64),
                ):
                    output[NewBall(sites, radius, shape=output.shape)] = 1
            else:
                output[
                    NewBall(
                        node.position,
                        np_.around(0.5 * (node.diameter - 1.0))
                        .astype(np_.int64)
                        .item(),
                        shape=output.shape,
                    )
                ] = 1

        for ___, ___, edge in self.edges.data("as_edge_t"):
            for *sites, radius in zip(
                *edge.sites, np_.around(0.5 * (edge.widths - 1.0)).astype(np_.int64)
            ):
                output[NewBall(sites, radius, shape=output.shape)] = 1

        return output

    def SetPlotStyles(
        self,
        /,
        *,
        node: Union[node_style_unstructured_h, node_styles_unstructured_h] = None,
        edge: edge_style_unstructured_h = None,
        regular_edge: edge_style_unstructured_h = None,
        self_loop: edge_style_unstructured_h = None,
        edges: edge_styles_unstructured_h = None,
        direction: direction_style_unstructured_h = None,
        label: label_style_unstructured_h = None,
        label_node: label_style_unstructured_h = None,
        label_edge: label_style_unstructured_h = None,
        labels: label_styles_unstructured_h = None,
    ) -> None:
        """
        edge: common style for both regulars and self-loops
        edges: styles for regulars and self-loops, in that order
        label: common style for both nodes and edges
        labels: styles for nodes and edges, in that order
        """
        if node is not None:
            if isinstance(node, tuple):  # node_style_unstructured_h
                self.node_styles = {
                    None: plot_style_t(type=node[0], size=node[1], color=node[2])
                }
            else:
                styles = {
                    _dgr: plot_style_t(type=_stl[0], size=_stl[1], color=_stl[2])
                    for _dgr, _stl in node.items()
                }
                self.node_styles.update(styles)
            if None not in self.node_styles:
                # Should never happen, but just in case...
                raise ValueError(
                    f"{self.node_styles}: Node styles without default style. "
                    f'Expected: style dictionary with a "None" entry.'
                )

        if edge is not None:
            style = plot_style_t(type=edge[0], size=edge[1], color=edge[2])
            styles = (style, style)
        elif regular_edge is not None:
            style = plot_style_t(
                type=regular_edge[0],
                size=regular_edge[1],
                color=regular_edge[2],
            )
            styles = (style, self.edge_styles[1])
        elif self_loop is not None:
            style = plot_style_t(
                type=self_loop[0],
                size=self_loop[1],
                color=self_loop[2],
            )
            styles = (self.edge_styles[0], style)
        elif edges is not None:
            styles = tuple(
                plot_style_t(
                    type=edges[_idx][0],
                    size=edges[_idx][1],
                    color=edges[_idx][2],
                )
                for _idx in range(2)
            )
        else:
            styles = None
        if styles is not None:
            self.edge_styles = styles

        if direction is not None:
            self.direction_style = direction_style_t(
                show=direction[0], size=direction[1], color=direction[2]
            )

        if label is not None:
            style = label_style_t(show=label[0], size=label[1], color=label[2])
            styles = (style, style)
        elif label_node is not None:
            style = label_style_t(
                show=label_node[0],
                size=label_node[1],
                color=label_node[2],
            )
            styles = (style, self.label_styles[1])
        elif label_edge is not None:
            style = label_style_t(
                show=label_edge[0],
                size=label_edge[1],
                color=label_edge[2],
            )
            styles = (self.label_styles[0], style)
        elif labels is not None:
            styles = tuple(
                label_style_t(
                    show=labels[_idx][0],
                    size=labels[_idx][1],
                    color=labels[_idx][2],
                )
                for _idx in range(2)
            )
        else:
            styles = None
        if styles is not None:
            self.label_styles = styles

    def Plot(
        self,
        figure: pl_.Figure = None,
        axes: pl_.axes.Axes = None,
        mode: plot_mode_e = plot_mode_e.SKL,
        max_distance: float = 1.0,
        colormap: Dict[int, str] = None,
        should_block: bool = True,
        should_return_figure: bool = False,
        should_return_axes: bool = False,
    ) -> Optional[Union[figure_t, axes_t, Tuple[figure_t, axes_t]]]:
        #
        if self.number_of_nodes() < 1:
            print(f"{__name__}.{self.Plot.__name__}: Empty graph")
            return None

        if axes is None:
            if figure is None:
                figure = pl_.figure()
            if self.dim == 2:
                axes = figure.gca()
            else:
                axes = figure.add_subplot(1, 1, 1, projection=m3_.Axes3D.name)
            axes.invert_yaxis()
        else:
            figure = axes.get_figure()

        if axes.yaxis_inverted():
            transformation = lambda y: y
            vector_transf = lambda y: y
        else:
            max_0 = self.domain_lengths[0] - 1
            transformation = lambda y: max_0 - np_.asarray(y)
            vector_transf = lambda y: -np_.asarray(y)

        transform_coords = lambda pos: (pos[1], transformation(pos[0]), *pos[2:])
        positions_as_dict = dict(
            (uid, transform_coords(node.position))
            for uid, node in self.nodes.data("as_node_t")
        )

        if self.dim == 2:
            if mode is plot_mode_e.Networkx:
                self._PlotWithNetworkX(positions_as_dict, axes, colormap)
            #
            elif mode in (
                plot_mode_e.SKL,
                plot_mode_e.SKL_Polyline,
                plot_mode_e.SKL_Curve,
            ):
                self._PlotExplicitly(
                    positions_as_dict,
                    transformation,
                    vector_transf,
                    axes,
                    mode,
                    max_distance,
                )
            #
            elif mode is plot_mode_e.Graphviz:
                self._PlotWithGraphviz(axes)
            #
            else:
                raise ValueError(f"{mode}: Invalid plotting mode")
            #
        else:
            self._PlotExplicitly(
                positions_as_dict,
                transformation,
                vector_transf,
                axes,
                mode,
                max_distance,
            )

        if self.dim == 2:
            # Matplotlib says: NotImplementedError: It is not currently possible to manually set the aspect on 3D axes
            axes.axis("equal")

        if should_block:
            pl_.show()  # Better named as TriggerMatplotlibEventLoop
            return None
        elif should_return_figure:
            if should_return_axes:
                return figure, axes
            else:
                return figure
        elif should_return_axes:
            return axes

        return None

    def _PlotWithNetworkX(
        self,
        positions_as_dict: Dict[str, Tuple[int, ...]],
        axes: pl_.axes.Axes,
        colormap: Optional[Dict[int, str]],
    ) -> None:
        #
        if colormap is None:
            colormap = self.__class__.colormap

        node_degrees = (elm[1] for elm in self.degree)
        node_colors = tuple(
            colormap[degree] if degree < 3 else colormap[2] for degree in node_degrees
        )

        nx_.draw_networkx(
            self,
            ax=axes,
            pos=positions_as_dict,
            node_color=node_colors,
            with_labels=self.label_styles[0].show,
            font_size=int(round(self.label_styles[0].size)),
            width=self.edge_styles[0].size,
        )
        if self.label_styles[1].show:
            nx_.draw_networkx_edge_labels(
                self,
                ax=axes,
                pos=positions_as_dict,
                edge_labels=self._EdgeIDsForPlot(),
                font_size=int(round(self.label_styles[1].size)),
            )

    def _PlotExplicitly(
        self,
        positions_as_dict: Dict[str, Tuple[int, ...]],
        transformation: Callable[[array_t], array_t],
        vector_transf: Callable[[array_t], array_t],
        axes: pl_.axes.Axes,
        mode: plot_mode_e,
        max_distance: float,
    ) -> None:
        #
        if mode == plot_mode_e.SKL:
            mode = "site"
        elif mode == plot_mode_e.SKL_Polyline:
            mode = "polyline"
        else:
            mode = "curve"
        dg_.Plot(
            self.edges.data("as_edge_t"),
            transformation,
            vector_transf,
            axes,
            self.edge_styles,
            self.direction_style,
            self.label_styles[1],
            mode=mode,
            max_distance=max_distance,
        )
        # TODO: there is no distinction between end nodes and isolated nodes
        nd_.PlotEndNodes(
            self.nodes.data("as_node_t"),
            transformation,
            axes,
            self.node_styles.get(1, self.node_styles[None]),
        )

        if self.dim == 2:
            nd_.Plot2DBranchNodes(
                self.nodes.data("as_node_t"),
                self.degree,
                transformation,
                axes,
                self.node_styles,
            )
            if self.label_styles[0].show:
                nx_.draw_networkx_labels(
                    self,
                    ax=axes,
                    pos=positions_as_dict,
                    font_size=int(round(self.label_styles[0].size)),
                    font_color=self.label_styles[0].color,
                )
        else:
            nd_.Plot3DBranchNodes(
                self.nodes.data("as_node_t"),
                self.degree,
                transformation,
                axes,
                self.node_styles,
            )
            if self.label_styles[0].show:
                nd_.Plot3DNodeLabels(self, positions_as_dict, axes, self.label_styles)

    def _PlotWithGraphviz(self, axes: pl_.axes.Axes) -> None:
        #
        try:
            import tempfile as tp_

            import imageio as io_
            import pygraphviz as gp_

            graph = nx_.nx_agraph.to_agraph(self)
            with tp_.NamedTemporaryFile() as tmp_accessor:
                img_name = tmp_accessor.name
                graph.layout()
                graph.draw(img_name, format="png")
                axes.imshow(io_.v3.imread(img_name))
        except Exception as exc:
            axes.text(
                0,
                0,
                f"Unable to plot graph using pygraphviz/imageio.\nPlease check installed modules.\n[{exc}]",
                horizontalalignment="center",
            )

    def _EdgeIDsForPlot(self) -> Dict[str, str]:
        #
        lengths_as_dict = nx_.get_edge_attributes(self, "length")
        w_lengths_as_dict = (
            nx_.get_edge_attributes(self, "w_length") if self.has_widths else None
        )

        w_length_str = ""
        edge_ids = {}
        for key, value in lengths_as_dict.items():
            if w_lengths_as_dict is not None:
                w_length_str = "/" + str(round(w_lengths_as_dict[key]))
            edge_ids[key[0:2]] = key[2] + "\n" + str(round(value)) + w_length_str

        return edge_ids

    def __str__(self) -> str:
        """"""
        output = (
            f"{self.__class__.__name__}:\n"
            f"    Domain lengths={self.domain_lengths}\n"
            f"    Has widths={self.has_widths}\n\n"
            f"    Components={self.n_components}\n"
            f"    Nodes={self.n_nodes}"
            f" = S_{self.n_s_nodes} + E_{self.n_e_nodes} + B_{self.n_b_nodes}\n"
            f"    Edges={self.n_edges}"
        )

        return output


def _Ball3D(center: Sequence[int], radius: int, shape: Tuple[int, int, int]) -> array_t:
    #
    output = np_.zeros(shape, dtype=np_.bool_)
    # dw_.ellipsoid leaves a one pixel margin around the ellipse, hence [1:-1, 1:-1, 1:-1]
    ellipse = dw_.ellipsoid(radius, radius, radius)[1:-1, 1:-1, 1:-1]
    sp_slices = tuple(
        slice(0, min(output.shape[idx_], ellipse.shape[idx_])) for idx_ in (0, 1, 2)
    )
    output[sp_slices] = ellipse[sp_slices]

    row, col, dep = center
    output = im_.shift(
        output, (row - radius, col - radius, dep - radius), order=0, prefilter=False
    )

    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# skl_ograph=Skeleton graph with specific operations, derived from skeleton graph with features

raise NotImplementedError("THIS MODULE IS A WORK-IN-PROGRESS: NOT USABLE YET")

# Derive an oedge_t (or other name) adding a field composition:
# composition: Holds the lengths of pieces composing the edge (can change while simplifying)

from skl_graph.skl_fgraph import skl_graph_t as skl_fgraph_t  # f=feature
# from skl_graph import EdgeID, NodeID


import networkx as nx_
import numpy as np_


array_t = np_.ndarray


class skl_graph_t(skl_fgraph_t):
    def PruneBasedOnWidths(self, min_width: float) -> None:
        #
        assert self.has_widths

        delete_list = []
        relabeling_dct = {}

        for node_0, node_1, edge_desc in self.edges.data("as_edge_t"):
            extremity = None
            if self.degree[node_0] == 1:
                extremity = node_0
            elif self.degree[node_1] == 1:
                extremity = node_1
            if extremity is None:
                continue

            assert (
                len(edge_desc["composition"]) == 1
            )  # /!\ Not made to work on already simplified graphs

            edge_coords_0 = edge_desc["sites"][0]
            edge_coords_1 = edge_desc["sites"][1]
            widths = edge_desc["widths"]
            n_edge_pixels = len(edge_coords_0)

            if (edge_coords_0[0], edge_coords_1[0]) == self.node[extremity].position:
                pixel_idx = 0
                idx_incr = 1
            else:
                pixel_idx = n_edge_pixels - 1
                idx_incr = -1

            while (0 <= pixel_idx < n_edge_pixels) and (widths[pixel_idx] < min_width):
                pixel_idx += idx_incr

            if (pixel_idx < 0) or (pixel_idx >= n_edge_pixels):
                delete_list.append(extremity)
            else:
                self.node[extremity].position = (
                    edge_coords_0[pixel_idx],
                    edge_coords_1[pixel_idx],
                )
                self.node[extremity].diameter = widths[pixel_idx]
                relabeling_dct[extremity] = NodeID(self.node[extremity].position)

                if idx_incr > 0:
                    valid_idc = slice(pixel_idx, n_edge_pixels)
                    extra_idc = slice(pixel_idx + 1)
                else:
                    valid_idc = slice(pixel_idx + 1)
                    extra_idc = slice(pixel_idx, n_edge_pixels)

                edge_piece = np_.array(
                    (edge_coords_0[extra_idc], edge_coords_1[extra_idc]), dtype=np_.float64
                )
                extra_widths = np_.array(widths[extra_idc], dtype=np_.float64)
                extra_lengths = np_.sqrt(
                    (np_.diff(edge_piece, axis=1) ** 2).sum(axis=0)
                )
                extra_w_lengths = extra_lengths * (
                    0.5 * (extra_widths[1:] + extra_widths[:-1])
                )

                edge_desc["sites"] = (
                    edge_coords_0[valid_idc],
                    edge_coords_1[valid_idc],
                )
                edge_desc["widths"] = widths[valid_idc]
                edge_desc[
                    "length"
                ] -= extra_lengths.sum().item()  # Conversion to float is necessary,
                edge_desc[
                    "w_length"
                ] -= (
                    extra_w_lengths.sum().item()
                )  # otherwise type np_.float64 contaminates all.
                if edge_desc["origin_node"] == extremity:
                    edge_desc["origin_node"] = relabeling_dct[extremity]

        if len(delete_list):
            self.remove_nodes_from(delete_list)
        if len(relabeling_dct):
            nx_.relabel_nodes(self, relabeling_dct, copy=False)

    def Simplify(self, min_edge_length: float) -> None:
        #
        while True:
            min_length = np_.Inf
            end_nodes = ()
            for node_label_0, node_label_1, edge in self.edges.data("as_edge_t"):
                degree_0 = self.degree[node_label_0]
                degree_1 = self.degree[node_label_1]

                if (degree_0 > 2) and (degree_1 > 2) and (edge.length < min_length):
                    edge_desc_list = self[node_label_0][node_label_1]
                    if len(edge_desc_list) > 1:
                        lengths_lower = [
                            description["length"] <= edge.length
                            for description in edge_desc_list.values()
                        ]
                        should_continue = all(lengths_lower)
                    else:
                        should_continue = True
                    if should_continue:
                        min_length = edge.length
                        end_nodes = (node_label_0, node_label_1)

            if min_length < min_edge_length:
                # /!\ management of edge descriptions is inexistant
                edges = []
                all_coords = [[], []]
                n_coords_per_piece = []
                for edge, description in self[end_nodes[0]][end_nodes[1]].items():
                    edges.append(edge)
                    sites = description["sites"]
                    all_coords[0].extend(sites[0])
                    all_coords[1].extend(sites[1])
                    n_coords_per_piece.append(len(sites[0]))

                all_coords = np_.array(all_coords, dtype=np_.float64)
                cum_n_coords_per_piece = np_.cumsum(n_coords_per_piece) - 1
                # Naming: actually "cumulative minus one" rather than "cumulative"

                centroid = all_coords.mean(axis=1, keepdims=True).round()
                closest_pixel_idx = (
                    ((all_coords - centroid) ** 2).sum(axis=0).argmin()
                )  # idx of first occurrence of min
                centroid = (
                    int(all_coords[0, closest_pixel_idx]),
                    int(all_coords[1, closest_pixel_idx]),
                )

                closest_edge_idx = cum_n_coords_per_piece.searchsorted(
                    closest_pixel_idx
                )
                shared_description = self[end_nodes[0]][end_nodes[1]][
                    edges[closest_edge_idx]
                ]
                # for edge in self[end_nodes[0]][end_nodes[1]].keys():
                #     if edge != edge_to_keep:
                #         self.remove_edge(end_nodes[0], end_nodes[1], key = edge)

                description = self.nodes[end_nodes[0]]
                description["position"] = centroid
                description["sites"] = ((centroid[0],), (centroid[1],))

                for neighbor in self[end_nodes[0]]:
                    if neighbor != end_nodes[1]:
                        for description in self[end_nodes[0]][neighbor].values():
                            pass
                            # use shared_description here
                            # description = {
                            #     "sites":           new_coords,
                            #     'origin_node': edge_0_node,
                            #     'length':           edge_0_desc['length']   + edge_1_desc['length']   + joint_length,
                            #     'w_length':         edge_0_desc['w_length'] + edge_1_desc['w_length'] + joint_w_length,
                            #     'widths':            widths,
                            #     'composition':      edge_0_desc['composition'] + (len(node_desc["sites"][0]),) + \
                            #                         edge_1_desc['composition']
                            # }

                for neighbor in self[end_nodes[1]]:
                    if (neighbor != end_nodes[0]) and (
                        neighbor not in self[end_nodes[0]]
                    ):
                        for edge, description in self[end_nodes[1]][neighbor].items():
                            # use shared_description here
                            # description = {
                            #     "sites":           new_coords,
                            #     'origin_node': edge_0_node,
                            #     'length':           edge_0_desc['length']   + edge_1_desc['length']   + joint_length,
                            #     'w_length':         edge_0_desc['w_length'] + edge_1_desc['w_length'] + joint_w_length,
                            #     'widths':            widths,
                            #     'composition':      edge_0_desc['composition'] + (len(node_desc["sites"][0]),) + \
                            #                         edge_1_desc['composition']
                            # }
                            self.add_edge(
                                end_nodes[0], neighbor, key=edge, **description
                            )

                self.remove_node(end_nodes[1])
            else:
                break

        while True:
            min_length = np_.Inf
            node_label = -1
            for node_label_0, node_label_1, edge in self.edges.data("as_edge_t"):
                degree_0 = self.degree[node_label_0]
                degree_1 = self.degree[node_label_1]

                if (
                    ((degree_0 == 1) or (degree_1 == 1))
                    and (degree_0 + degree_1 > 3)
                    and (edge.length < min_length)
                ):
                    min_length = edge.length
                    if degree_0 == 1:
                        node_label = node_label_0
                    else:
                        node_label = node_label_1

            if min_length < min_edge_length:
                self.remove_node(node_label)
            else:
                break

        graph_has_been_modified = True
        while graph_has_been_modified:
            graph_has_been_modified = False

            for node_label, node_desc in self.nodes.data("as_node_t"):
                if self.degree[node_label] != 2:
                    continue

                edge_0, edge_1 = self.edges(node_label, data=True)
                other_node_0, edge_0_node, edge_0_desc = edge_0
                other_node_1, edge_1_node, edge_1_desc = edge_1
                assert (other_node_0 == node_label) and (other_node_1 == node_label)
                # If this assertion fails one day, it means that NetworkX has changed the way it returns
                # adjacent edges. It will become necessary to test which of other_node_X and edge_X_node
                # is node_label.

                new_coords, joint_length, first_reversed, last_reversed = __EdgeOfGluedEdges__(
                    edge_0_desc["sites"],
                    edge_1_desc["sites"],
                    node_desc["sites"],
                    edge_0_desc["origin_node"] == node_label,
                    edge_1_desc["origin_node"] == node_label,
                )

                if self.has_widths:
                    joint_w_length = joint_length * np_.mean(node_desc["diameters"])
                    if first_reversed:
                        first_part = tuple(reversed(edge_0_desc["widths"]))
                    else:
                        first_part = edge_0_desc["widths"]
                    if last_reversed:
                        last_part = tuple(reversed(edge_1_desc["widths"]))
                    else:
                        last_part = edge_1_desc["widths"]
                    widths = first_part + node_desc["diameters"] + last_part
                else:
                    joint_w_length = 0
                    widths = None

                description = {
                    "sites": new_coords,
                    "origin_node": edge_0_node,
                    "length": edge_0_desc["length"]
                    + edge_1_desc["length"]
                    + joint_length,
                    "w_length": edge_0_desc["w_length"]
                    + edge_1_desc["w_length"]
                    + joint_w_length,
                    "widths": widths,
                    "composition": edge_0_desc["composition"]
                    + (len(node_desc["sites"][0]),)
                    + edge_1_desc["composition"],
                }

                self.add_edge(
                    edge_0_node,
                    edge_1_node,
                    EdgeID(edge_0_node, edge_1_node),
                    **description
                )
                self.remove_node(node_label)
                graph_has_been_modified = True
                break


def __EdgeOfGluedEdges__(
    edge_0_coords, edge_1_coords, node_coords, node_is_first_of_0, node_is_first_of_1
):
    #
    # Returns the glued sites and the length of the gluing joint
    #
    if node_is_first_of_0:
        first_part_0 = tuple(reversed(edge_0_coords[0]))
        first_part_1 = tuple(reversed(edge_0_coords[1]))
        first_reversed = True
    else:
        first_part_0 = edge_0_coords[0]
        first_part_1 = edge_0_coords[1]
        first_reversed = False

    if node_is_first_of_1:
        last_part_0 = tuple(reversed(edge_1_coords[0]))
        last_part_1 = tuple(reversed(edge_1_coords[1]))
        last_reversed = True
    else:
        last_part_0 = edge_1_coords[0]
        last_part_1 = edge_1_coords[1]
        last_reversed = False

    glued_coords = (
        first_part_0 + node_coords[0] + last_part_0,
        first_part_1 + node_coords[1] + last_part_1,
    )
    joint_length = float(
        np_.sqrt(
            (first_part_0[-1] - last_part_0[0]) ** 2
            + (first_part_1[-1] - last_part_1[0]) ** 2
        )
    )

    return glued_coords, joint_length, first_reversed, last_reversed
